//
//  MBAppleScriptApplication.m
//  iTunesInterfaceTest
//
//  Created by Milo on 27/12/2006.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "MBAppleScriptApplication.h"
#import "MBAppleScriptApplicationPrivate.h"
#import "MBAppleScriptObject.h"
#import "MBAppleScriptObjectPrivate.h"
#import "MBType.h"
#import "MBProcessSerialNumber.h"
#import "DLog.h"

NSString *MBAppleScriptApplicationDidLaunchNotification = @"MBAppleScriptApplicationDidLaunchNotification";
NSString *MBAppleScriptApplicationDidTerminateNotification = @"MBAppleScriptApplicationDidTerminateNotification";


@implementation MBAppleScriptApplication

+ (id)applicationWithName:(NSString *)name shouldUseCacheDefault:(BOOL)shouldUseCacheDefault;
{
	return [[[self alloc] initWithName:name shouldUseCacheDefault:shouldUseCacheDefault] autorelease];
}

- (id)initWithName:(NSString *)name shouldUseCacheDefault:(BOOL)shouldUseCacheDefault;
{
	if ([super _initWithObjectSpecifier:[NSAppleEventDescriptor nullDescriptor] application:self shouldUseCache:shouldUseCacheDefault] == nil)
		return nil;
	
	if (name == nil)
	{
		[[NSException exceptionWithName:NSInvalidArgumentException reason:nil userInfo:nil] raise];
		[self release];
		return nil;
	}
	
	[self _setName:name];
	[self setShouldUseCacheDefault:shouldUseCacheDefault];
	
	[[[NSWorkspace sharedWorkspace] notificationCenter] addObserver:self selector:@selector(_workspaceDidLaunchApplication:) name:NSWorkspaceDidLaunchApplicationNotification object:nil];
	[[[NSWorkspace sharedWorkspace] notificationCenter] addObserver:self selector:@selector(_workspaceDidTerminateApplication:) name:NSWorkspaceDidTerminateApplicationNotification object:nil];
	
	return self;
}

// MBAppleScriptObject
- (id)_initWithObjectSpecifier:(NSAppleEventDescriptor *)objectSpecifier application:(MBAppleScriptApplication *)application shouldUseCache:(BOOL)shouldUseCache;
{
	return [self initWithName:nil shouldUseCacheDefault:YES];	// nil
}

// MBAppleScriptObject
- (void)dealloc
{
	[self _setName:nil];
	[self _setProcessSerialNumber:nil];
	
	[super dealloc];
}

// MBAppleScriptObject
- (NSString *)description
{
	return [self name];
}

- (NSString *)name;
{
	return _name;
}

- (NSBundle *)bundle;
{
	return [NSBundle bundleWithPath:[[NSWorkspace sharedWorkspace] fullPathForApplication:[self name]]];	// nil if the application can't be found
}

- (BOOL)shouldUseCacheDefault;
{
	return _shouldUseCacheDefault;
}

- (void)setShouldUseCacheDefault:(BOOL)newDefault;
{
	_shouldUseCacheDefault = newDefault;
}

@end


#pragma mark -
@implementation MBAppleScriptApplication (Private)

- (MBProcessSerialNumber *)_processSerialNumber;
{
	MBProcessSerialNumber *PSN;
	@synchronized(_processSerialNumber)
	{
		PSN = [_processSerialNumber retain];
	}
	[PSN autorelease];
	
	if (PSN == nil)
	{
		NSArray *runningApplications = [[NSWorkspace sharedWorkspace] launchedApplications];
		NSEnumerator *enumerator = [runningApplications objectEnumerator];
		NSDictionary *someApplication;
		while (someApplication = [enumerator nextObject])
		{
			if ([[someApplication objectForKey:@"NSApplicationBundleIdentifier"] isEqualToString:[[self bundle] bundleIdentifier]])
			{
				PSN = [[[MBProcessSerialNumber alloc] initWithHighLong:[someApplication objectForKey:@"NSApplicationProcessSerialNumberHigh"]
															   lowLong:[someApplication objectForKey:@"NSApplicationProcessSerialNumberLow"]] autorelease];
				[self _setProcessSerialNumber:PSN];
			}
		}
	}
	
	return PSN;
}

- (void)_setProcessSerialNumber:(MBProcessSerialNumber *)newPSN;
{
	MBProcessSerialNumber *originalValue;
	[newPSN retain];
	@synchronized(_processSerialNumber)
	{
		originalValue = _processSerialNumber;
		_processSerialNumber = newPSN;
	}
	[originalValue release];
}

- (void)_setName:(NSString *)newName;
{
	if (newName == _name)
		return;
	[_name release];
	_name = [newName retain];
}

- (NSAppleEventDescriptor *)_targetDescriptor;
{
	return [[self _launchAndHide] targetDescriptor];
}

- (MBProcessSerialNumber *)_launchAndHide;
{
	MBProcessSerialNumber *processSerialNumber;
	if (processSerialNumber = [self _processSerialNumber])		// application already running
		return processSerialNumber;
		
	else if ([self bundle] == nil)
	{
		DLog(@"%s - %@ not found!", _cmd, [self name]);
		return nil;
	}
	
	else if ([[NSWorkspace sharedWorkspace] launchAppWithBundleIdentifier:[[self bundle] bundleIdentifier] 
																  options:NSWorkspaceLaunchAndHide additionalEventParamDescriptor:nil 
														 launchIdentifier:nil] == NO)
	{
		DLog(@"%s - %@ launch failed!", _cmd, [self name]);
		return nil;
	}
	
	else
	{
		DLog(@"%s - %@ launched successfully", _cmd, [self name]);
		return [self _processSerialNumber];
	}	
}

- (void)_workspaceDidLaunchApplication:(NSNotification *)notification;
{
	if ([[[notification userInfo] objectForKey:@"NSApplicationName"] isEqualToString:[self name]] == NO)
		return;
	
	[[NSNotificationCenter defaultCenter] postNotificationName:MBAppleScriptApplicationDidLaunchNotification object:self];
}

- (void)_workspaceDidTerminateApplication:(NSNotification *)notification;
{
	if ([[[notification userInfo] objectForKey:@"NSApplicationName"] isEqualToString:[self name]] == NO)
		return;
	
	[self _setProcessSerialNumber:nil];
	[[NSNotificationCenter defaultCenter] postNotificationName:MBAppleScriptApplicationDidTerminateNotification object:self];
}

@end