//
//  MBAppleScriptObject.h
//  MBAppleScriptBridge
//
//	A class to represent an object in an application's Apple Script dictionary.
//	You should not attempt to allocate instances of this class yourself. Instead, create an MBAppleScriptApplication instance and use that to get associated instances of MBAppleScriptObject.
//	Thread safe.
//
//  Created by Milo on 27/12/2006.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import <Cocoa/Cocoa.h>
@class MBAppleScriptApplication;


@interface MBAppleScriptObject : NSObject 
{
	MBAppleScriptApplication *_application;
	NSAppleEventDescriptor *_objectSpecifier;
	NSMutableDictionary *_cache;
	BOOL _shouldUseCache;
}

- (MBAppleScriptApplication *)application;

- (void)resetCache;									// Empties the cache
- (void)resetCachedValueForType:(OSType)type;		// Removes a property or element from the cache

- (BOOL)shouldUseCache;
- (void)setShouldUseCache:(BOOL)newShouldUseCache;

// AppleScript Commands
- (id)performCommandWithEventClass:(OSType)eventClass eventID:(OSType)eventID arguments:(NSDictionary *)argumentsDictionary;	// The receiver is used as the direct object, argumentsDictionary can be nil

// AppleScript Elements
- (NSArray *)elementsWithType:(OSType)classType properties:(NSDictionary *)propertiesDictionary;		// propertiesDictionary can be nil

// AppleScript Properties
- (id)propertyWithType:(OSType)propertyType;
- (BOOL)setPropertyWithType:(OSType)propertyType toValue:(id)newPropertyValue;
@end
