//
//  MBAppleScriptObject.m
//  iTunesInterfaceTest
//
//  Created by Milo on 27/12/2006.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "MBAppleScriptObject.h"
#import "MBAppleScriptObjectPrivate.h"
#import "MBAppleScriptApplication.h"
#import "MBAppleScriptApplicationPrivate.h"
#import "MBType.h"
#import "DLog.h"


@implementation MBAppleScriptObject

// NSObject
- (id)init
{
	return [self _initWithObjectSpecifier:nil application:nil shouldUseCache:YES];		// nil
}

// NSObject
- (void)dealloc
{
	[self _setObjectSpecifier:nil];
	[self _setApplication:nil];
	[_cache release];
	
	[[[NSWorkspace sharedWorkspace] notificationCenter] removeObserver: self];
	
	[super dealloc];
}

// NSObject
- (NSString *)description
{
	return [NSString stringWithFormat:@"%@ - %@", [self application], [self _objectSpecifier]];
}

- (MBAppleScriptApplication *)application;
{
	return _application;
}

- (void)resetCache;
{
	if ([self shouldUseCache] == NO)
		return;
	@synchronized(_cache)
	{
		[_cache removeAllObjects];
	}
}

- (void)resetCachedValueForType:(OSType)type;
{
	[self _setCachedValue:nil forType:type];
}

- (BOOL)shouldUseCache;
{
	return _shouldUseCache;
}

- (void)setShouldUseCache:(BOOL)newShouldUseCache;
{
	if (newShouldUseCache == NO)
		[self resetCache];		// Resets the cache if it is being switched off (and was previously on!)
	_shouldUseCache = newShouldUseCache;
}

#pragma mark -
#pragma mark AppleScript

- (id)performCommandWithEventClass:(OSType)eventClass eventID:(OSType)eventID arguments:(NSDictionary *)argumentsDictionary;
{
	NSAppleEventDescriptor *appleEvent = [NSAppleEventDescriptor appleEventWithTargetDescriptor:[[self application] _targetDescriptor] eventClass:eventClass eventID:eventID];
	if (appleEvent == nil)
		return nil;
	[appleEvent setDescriptor:[self _objectSpecifier] forKeyword:keyDirectObject];
	if (argumentsDictionary != nil)
	{
		NSEnumerator *keyEnumerator = [argumentsDictionary keyEnumerator];
		MBType *dictionaryKey;
		while (dictionaryKey = [keyEnumerator nextObject])
		{
			if ([dictionaryKey isKindOfClass:[MBType class]])
				[appleEvent setDescriptor:[NSAppleEventDescriptor descriptorWithCocoaEquivalent:[argumentsDictionary objectForKey:dictionaryKey]] forKeyword:[dictionaryKey OSType]];
		}
	}
	
	appleEvent = [appleEvent sendWithImmediateReply];
	if (appleEvent == nil)
		[[self application] _setProcessSerialNumber:nil];	// Reset the process serial number if sending fails
	
	return [[appleEvent directObject] cocoaEquivalent:self];
}

- (NSArray *)elementsWithType:(OSType)classType properties:(NSDictionary *)propertiesDictionary;
{		
	if (propertiesDictionary == nil || [propertiesDictionary count] == 0)	// We won't use the cache if properties have been specified
	{
		id cachedValue = [self _cachedValueForType:classType];
		if (cachedValue != nil)
			return cachedValue;
	}
	
	NSAppleEventDescriptor *appleEvent = [NSAppleEventDescriptor appleEventWithTargetDescriptor:[[self application] _targetDescriptor] eventClass:kAECoreSuite eventID:kAEGetData];
	if (appleEvent == nil)
		return nil;
	NSAppleEventDescriptor *objectRecord = [[NSAppleEventDescriptor recordDescriptor] coerceToDescriptorType:typeObjectSpecifier];
	{
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithEnumCode:formAbsolutePosition] forKeyword:keyAEKeyForm];
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:classType] forKeyword:keyAEDesiredClass];
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithAbsoCode:kAEAll] forKeyword:keyAEKeyData];
		if (propertiesDictionary == nil || [propertiesDictionary count] == 0)
			[objectRecord setDescriptor:[self _objectSpecifier] forKeyword:keyAEContainer];
		else
		{
			NSAppleEventDescriptor *fromObjectRecord = [[NSAppleEventDescriptor recordDescriptor] coerceToDescriptorType:typeObjectSpecifier];
			{
				[fromObjectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithEnumCode:formTest] forKeyword:keyAEKeyForm];
				[fromObjectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:classType] forKeyword:keyAEDesiredClass];
				[fromObjectRecord setDescriptor:[self _objectSpecifier] forKeyword:keyAEContainer];
				NSAppleEventDescriptor *logicRecord = [[NSAppleEventDescriptor recordDescriptor] coerceToDescriptorType:typeLogicalDescriptor];
				{
					[logicRecord setDescriptor:[NSAppleEventDescriptor descriptorWithEnumCode:kAEAND] forKeyword:keyAELogicalOperator];
					NSAppleEventDescriptor *termsDescriptor = [NSAppleEventDescriptor listDescriptor];
					{
						NSEnumerator *keyEnumerator = [propertiesDictionary keyEnumerator];
						MBType *dictionaryKey;
						while (dictionaryKey = [keyEnumerator nextObject])
						{
							if ([dictionaryKey isKindOfClass:[MBType class]])
							{
								NSAppleEventDescriptor *comparisonRecord = [[NSAppleEventDescriptor recordDescriptor] coerceToDescriptorType:typeCompDescriptor];
								{
									[comparisonRecord setDescriptor:[NSAppleEventDescriptor descriptorWithEnumCode:kAEEquals] forKeyword:keyAECompOperator];
									NSAppleEventDescriptor *object1Record = [[NSAppleEventDescriptor recordDescriptor] coerceToDescriptorType:typeObjectSpecifier];
									{
										[object1Record setDescriptor:[NSAppleEventDescriptor descriptorWithEnumCode:typeProperty] forKeyword:keyAEKeyForm];
										[object1Record setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:typeProperty] forKeyword:keyAEDesiredClass];
										[object1Record setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:[dictionaryKey OSType]] forKeyword:keyAEKeyData];
										[object1Record setDescriptor:[NSAppleEventDescriptor descriptorWithExmnType] forKeyword:keyAEContainer];
									}
									[comparisonRecord setDescriptor:object1Record forKeyword:keyAEObject1];
									[comparisonRecord setDescriptor:[NSAppleEventDescriptor descriptorWithCocoaEquivalent:[propertiesDictionary objectForKey:dictionaryKey]] forKeyword:keyAEObject2];
								}
								[termsDescriptor insertDescriptor:comparisonRecord atIndex:0];
							}
						}
					}
					[logicRecord setDescriptor:termsDescriptor forKeyword:keyAELogicalTerms];
				}
				[fromObjectRecord setDescriptor:logicRecord forKeyword:keyAEKeyData];
			}
			[objectRecord setDescriptor:fromObjectRecord forKeyword:keyAEContainer];
		}
	}
	[appleEvent setDescriptor:objectRecord forKeyword:keyDirectObject];	
	
	appleEvent = [appleEvent sendWithImmediateReply];
	if (appleEvent == nil)
		[[self application] _setProcessSerialNumber:nil];	// Reset the process serial number if sending fails
	
	NSArray *replyArray = [[appleEvent directObject] cocoaEquivalent:self];
	if (replyArray != nil && [replyArray isKindOfClass:[NSArray class]] == YES)
	{
		if (propertiesDictionary == nil || [propertiesDictionary count] == 0)
			[self _setCachedValue:replyArray forType:classType];
		return replyArray;
	}
	else return nil;
}

- (id)propertyWithType:(OSType)propertyType;
{
	id cachedValue = [self _cachedValueForType:propertyType];
	if (cachedValue != nil)
		return cachedValue;
	
	NSAppleEventDescriptor *appleEvent = [NSAppleEventDescriptor appleEventWithTargetDescriptor:[[self application] _targetDescriptor] eventClass:kAECoreSuite eventID:kAEGetData];
	if (appleEvent == nil)
		return nil;
	NSAppleEventDescriptor *objectRecord = [[NSAppleEventDescriptor recordDescriptor] coerceToDescriptorType:typeObjectSpecifier];
	{
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithEnumCode:typeProperty] forKeyword:keyAEKeyForm];
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:typeProperty] forKeyword:keyAEDesiredClass];
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:propertyType] forKeyword:keyAEKeyData];
		[objectRecord setDescriptor:[self _objectSpecifier] forKeyword:keyAEContainer];
	}
	[appleEvent setDescriptor:objectRecord forKeyword:keyDirectObject];
			
	appleEvent = [appleEvent sendWithImmediateReply];
	if (appleEvent == nil)
		[[self application] _setProcessSerialNumber:nil];	// Reset the process serial number if sending fails
	
	id replyObject = [[appleEvent directObject] cocoaEquivalent:self];
	if (replyObject != nil)
		[self _setCachedValue:replyObject forType:propertyType];
	return replyObject;
}

- (BOOL)setPropertyWithType:(OSType)propertyType toValue:(id)newPropertyValue;
{
	if (newPropertyValue == nil)
		return NO;

	NSAppleEventDescriptor *appleEvent = [NSAppleEventDescriptor appleEventWithTargetDescriptor:[[self application] _targetDescriptor] eventClass:kAECoreSuite eventID:kAESetData];
	if (appleEvent == nil)
		return nil;
	[appleEvent setDescriptor:[NSAppleEventDescriptor descriptorWithCocoaEquivalent:newPropertyValue] forKeyword:keyAEData];
	NSAppleEventDescriptor *objectRecord = [[NSAppleEventDescriptor recordDescriptor] coerceToDescriptorType:typeObjectSpecifier];
	{
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithEnumCode:typeProperty] forKeyword:keyAEKeyForm];
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:typeProperty] forKeyword:keyAEDesiredClass];
		[objectRecord setDescriptor:[NSAppleEventDescriptor descriptorWithTypeCode:propertyType] forKeyword:keyAEKeyData];
		[objectRecord setDescriptor:[self _objectSpecifier] forKeyword:keyAEContainer];
	}
	[appleEvent setDescriptor:objectRecord forKeyword:keyDirectObject];	

	appleEvent = [appleEvent sendWithImmediateReply];
	if (appleEvent == nil)
		[[self application] _setProcessSerialNumber:nil];	// Reset the process serial number if sending fails

	if ([appleEvent directObject] != nil)
	{
		[self resetCachedValueForType:propertyType];
		return YES;
	}
	else return NO;
}

@end


#pragma mark -
@implementation MBAppleScriptObject (Private)

- (id)_initWithObjectSpecifier:(NSAppleEventDescriptor *)objectSpecifier application:(MBAppleScriptApplication *)application shouldUseCache:(BOOL)shouldUseCache;
{
	if ([super init] == nil)
		return nil;
		
	if (objectSpecifier == nil || application == nil)
	{
		[[NSException exceptionWithName:NSInvalidArgumentException reason:nil userInfo:nil] raise];
		[self release];
		return nil;
	}
		
	[self _setObjectSpecifier:objectSpecifier];
	[self _setApplication:application];
	[self setShouldUseCache:shouldUseCache];
	_cache = [[NSMutableDictionary alloc] initWithCapacity:([self shouldUseCache] ? 20 : 0)];	// If the cache is on, allocate some space in the cache for best performance
	
	[[[NSWorkspace sharedWorkspace] notificationCenter] addObserver:self selector:@selector(_appleScriptApplicationDidTerminate:) name:MBAppleScriptApplicationDidTerminateNotification object:[self application]];
	
	return self;
}

- (NSAppleEventDescriptor *)_objectSpecifier;
{
	return _objectSpecifier;
}

- (void)_setObjectSpecifier:(NSAppleEventDescriptor *)newObjectSpecifier;
{
	if (newObjectSpecifier == _objectSpecifier)
		return;
	[_objectSpecifier release];
	_objectSpecifier = [newObjectSpecifier retain];
}

- (void)_setApplication:(MBAppleScriptApplication *)newApplication;
{
	if (newApplication != self)
		[newApplication retain];
	if (_application != self)
		[_application release];
	_application = newApplication;
}

- (id)_cachedValueForType:(OSType)type;
{
	if ([self shouldUseCache] == NO)
		return nil;
	MBType *typeKey = [MBType typeWithOSType:type];
	id value;
	@synchronized(_cache)
	{
		value = [[_cache objectForKey:typeKey] retain];
	}
	return [value autorelease];
}

- (void)_setCachedValue:(id)value forType:(OSType)type;
{
	if ([self shouldUseCache] == NO)
		return;
	MBType *typeKey = [MBType typeWithOSType:type];
	if (value == nil)
	{
		@synchronized(_cache)
		{
			[_cache removeObjectForKey:typeKey];
		}
	}
	else
	{
		@synchronized(_cache)
		{
			[_cache setObject:value forKey:typeKey];
		}
	}
}

- (void)appleScriptApplicationDidTerminate:(NSNotification *)notification;
{
	[self resetCache];
}

@end


#pragma mark -
@implementation NSAppleEventDescriptor (MBAppleScriptObject)

+ (NSAppleEventDescriptor *)appleEventWithTargetDescriptor:(NSAppleEventDescriptor *)targetDescriptor eventClass:(AEEventClass)eventClass eventID:(AEEventID)eventID;
{
	if (targetDescriptor == nil)
		return nil;
	
	else return [NSAppleEventDescriptor appleEventWithEventClass:eventClass
														 eventID:eventID
												targetDescriptor:targetDescriptor
														returnID:kAutoGenerateReturnID
												   transactionID:kAnyTransactionID];
}

+ (NSAppleEventDescriptor *)descriptorWithExmnType;
{
	return [NSAppleEventDescriptor descriptorWithDescriptorType:typeObjectBeingExamined bytes:NULL length:0];
}

+ (NSAppleEventDescriptor *)descriptorWithAbsoCode:(OSType)ordinalCode;
{
	return [NSAppleEventDescriptor descriptorWithDescriptorType:typeAbsoluteOrdinal bytes:&ordinalCode length:sizeof(ordinalCode)];
}

+ (NSAppleEventDescriptor *)descriptorWithCocoaEquivalent:(id)object;
{
	if (object == nil)
		return nil;
	
	else if ([object isKindOfClass:[NSNull class]])
		return [NSAppleEventDescriptor nullDescriptor];
		
	else if ([object isKindOfClass:[MBAppleScriptObject class]])
		return [object _objectSpecifier];
	
	else if ([object isKindOfClass:[MBType class]])
		return [NSAppleEventDescriptor descriptorWithTypeCode:[object OSType]];
	
	else if ([object isKindOfClass:[NSDictionary class]])
	{
		NSAppleEventDescriptor *recordDescriptor = [NSAppleEventDescriptor recordDescriptor];
		NSEnumerator *keyEnumerator = [object keyEnumerator];
		MBType *dictionaryKey;
		while (dictionaryKey = [keyEnumerator nextObject])
		{
			if ([dictionaryKey isKindOfClass:[MBType class]])
				[recordDescriptor setDescriptor:[NSAppleEventDescriptor descriptorWithCocoaEquivalent:[object objectForKey:dictionaryKey]] forKeyword:[dictionaryKey OSType]];
		}
		return recordDescriptor;
	}
	
	else if ([object isKindOfClass:[NSArray class]])
	{
		NSAppleEventDescriptor *listDescriptor = [NSAppleEventDescriptor listDescriptor];
		NSEnumerator *enumerator = [object objectEnumerator];
		id arrayObject;
		while (arrayObject = [enumerator nextObject])
		{
			[listDescriptor insertDescriptor:[NSAppleEventDescriptor descriptorWithCocoaEquivalent:arrayObject] atIndex:0];
		}
		return listDescriptor;
	}
	
	else if ([object isKindOfClass:[NSString class]])
	{
		if ([[NSFileManager defaultManager] fileExistsAtPath:[object stringByExpandingTildeInPath]])	// String is a path
		{
			FSRef fileRef;
			OSErr error = FSPathMakeRef((UInt8 *)[object fileSystemRepresentation], &fileRef, NULL);
			if (error != noErr)
				return nil;
			AliasHandle fileAlias;
			error = FSNewAliasMinimal(&fileRef, &fileAlias);
			if (error != noErr)
				return nil;
			return [NSAppleEventDescriptor descriptorWithDescriptorType:typeAlias bytes:*fileAlias length:GetHandleSize((Handle)fileAlias)];
		}
		else	// String is probably just text
		{
			return [NSAppleEventDescriptor descriptorWithString:object];
		}
	}
	
	else if ([object isKindOfClass:[NSNumber class]])
		return [NSAppleEventDescriptor descriptorWithInt32:[object longValue]];
	
	else if ([object isKindOfClass:[NSData class]])
		return [NSAppleEventDescriptor descriptorWithDescriptorType:typeData data:object];
		
	//else if ([object isKindOfClass:[NSImage class]])
	//	return [NSAppleEventDescriptor descriptorWithDescriptorType:typePict data:[object data]];
	
	else if ([object isKindOfClass:[NSAppleEventDescriptor class]])
		return object;			// Return NSAppleEventDescriptor objects unchanged
	
	else return nil;	
}

- (id)cocoaEquivalent:(MBAppleScriptObject *)sender;
{
	if ([self descriptorType] == typeNull)
		return [NSNull null];
	
	else if ([self descriptorType] == typeObjectSpecifier)
		return [[[MBAppleScriptObject alloc] _initWithObjectSpecifier:self application:[sender application] shouldUseCache:[[sender application] shouldUseCacheDefault]] autorelease];
	
	else if ([self descriptorType] == typeEnumerated)
		return [MBType typeWithOSType:[self enumCodeValue]];
	
	else if ([self descriptorType] == typeType)
		return [MBType typeWithOSType:[self typeCodeValue]];
	
	else if ([self descriptorType] == typeAERecord)
	{
		NSMutableDictionary *descriptorDictionary = [[[NSMutableDictionary alloc] initWithCapacity:[self numberOfItems]] autorelease];
		short int i;
		for (i = 1; i <= [self numberOfItems]; i++)
			[descriptorDictionary setObject:[[self descriptorForKeyword:[self keywordForDescriptorAtIndex:i]] cocoaEquivalent:sender] forKey:[MBType typeWithOSType:[self keywordForDescriptorAtIndex:i]]];
		return [[descriptorDictionary copy] autorelease];
	}
	
	else if ([self descriptorType] == typeAEList)
	{
		NSMutableArray *descriptorArray = [[[NSMutableArray alloc] initWithCapacity:[self numberOfItems]] autorelease];
		short int i;
		for (i = 1; i <= [self numberOfItems]; i++)
			[descriptorArray addObject:[[self descriptorAtIndex:i] cocoaEquivalent:sender]];
		return [[descriptorArray copy] autorelease];
	}
	
	else if ([self descriptorType] == typeText || [self descriptorType] == typeUnicodeText)
		return [self stringValue];
	
	else if ([self descriptorType] == typeAlias)
	{
		NSData *descriptorData = [self data];
		AliasHandle fileAlias = (AliasHandle)NewHandle([descriptorData length]);
		if (fileAlias == NULL)
			return nil;
		[descriptorData getBytes:*fileAlias];
		
		FSRef fileRef;
		Boolean wasChanged;
		OSErr error = FSResolveAliasWithMountFlags(NULL, fileAlias, &fileRef, &wasChanged, kResolveAliasFileNoUI);
		DisposeHandle((Handle)fileAlias);
		if (error != noErr)
			return nil;
		
		CFURLRef resolvedURL = CFURLCreateFromFSRef(NULL, &fileRef);
		if (resolvedURL == NULL)
			return nil;
		
		NSString *resolvedPath = (NSString *)CFURLCopyFileSystemPath(resolvedURL, kCFURLPOSIXPathStyle);
		CFRelease(resolvedURL);
		return [resolvedPath autorelease];
	}
	
	else if ([self descriptorType] == typeSInt32)
		return [NSNumber numberWithLong:[self int32Value]];
	
	else if ([self descriptorType] == typeSInt64)
	{
		long long longLongBuffer;
		[[self data] getBytes:&longLongBuffer];
		return [NSNumber numberWithLongLong:longLongBuffer];
	}
	
	else if ([self descriptorType] == typeIEEE32BitFloatingPoint)
	{
		float floatBuffer;
		[[self data] getBytes:&floatBuffer];
		return [NSNumber numberWithFloat:floatBuffer];
	}
	
	else if ([self descriptorType] == typeData)
		return [self data];
	
	else if ([self descriptorType] == typePict)
		return [[[NSImage alloc] initWithData:[self data]] autorelease];
	
	else return self;	// Return an NSAppleEventDescriptor if we don't know how to translate the descriptor's value
}

- (NSAppleEventDescriptor *)sendWithImmediateReply;
{
	DLog(@"Sending Apple Event...");
	AppleEvent reply;
	OSStatus error = AESendMessage([self aeDesc], &reply, kAEWaitReply, kAEDefaultTimeout);
	if (error != noErr)
	{
		DLog(@"Error calling AESendMessage(): %d", error);
		return nil;
	}
	else return [[[NSAppleEventDescriptor alloc] initWithAEDescNoCopy:&reply] autorelease];	
}

- (NSAppleEventDescriptor *)directObject;
{
	NSAppleEventDescriptor *contentsDescriptor;
	if (contentsDescriptor = [self descriptorForKeyword:keyDirectObject])	// Reply event contains a direct object
	{
		return contentsDescriptor;
	}
	else if (contentsDescriptor = [self descriptorForKeyword:keyErrorNumber])	// Reply event contains an error number
	{
		DLog(@"Reply contains error from application: %i", [contentsDescriptor int32Value]);
		return nil;
	}
	else if (contentsDescriptor = [self descriptorForKeyword:keyErrorString])	// Reply event contains an error string
	{
		DLog(@"Reply contains error from appliation: %@", [contentsDescriptor stringValue]);
		return nil;
	}
	else return [NSAppleEventDescriptor nullDescriptor];	// Reply is empty or indeterminate
}

@end
